﻿using System;
using System.Collections.Generic;
using System.Net;
using System.Threading;
using System.Windows.Forms;

namespace PWSPlugin
{
    public class Plugin : IPlugin
    {
        private Dictionary<Guid, PluginInstance> instances;
        private Thread uiThread;
        private volatile bool isOpeningSettingDialog = false;

        private Thread clientThread;
        private volatile bool isClientRunning = false;
        private AutoResetEvent clientStopped;
        private ManualResetEvent clientFailed;
        private P2PQRelay.Client client = null;

        public Plugin()
        {
            this.instances = new Dictionary<Guid, PluginInstance>();

            if(P2PQRelayClient_Net.Properties.Settings.Default.IsUpgraded)
            {
                P2PQRelayClient_Net.Properties.Settings.Default.Upgrade();
                P2PQRelayClient_Net.Properties.Settings.Default.Save();
            }

            StartClient();
        }

        public IPluginInstance CreateNewInstance()
        {
            PluginInstance inst = new PluginInstance(this);
            this.instances.Add(inst.Guid, inst);

            return inst;
        }

        public void DestroyInstance(PluginInstance inst)
        {
            this.instances.Remove(inst.Guid);
        }

        public void Destroy()
        {
            this.instances.Clear();
            StopClient();
        }

        public void ShowSettingDialog()
        {
            if (!isOpeningSettingDialog)
            {
                isOpeningSettingDialog = true;

                this.uiThread = new Thread((ThreadStart)(() =>
                {
                    Action onClosed = new Action(this.OnSettingApplying);
                    Application.EnableVisualStyles();
                    Application.SetCompatibleTextRenderingDefault(false);
                    P2PQRelayClient_Net.SettingForm f = new P2PQRelayClient_Net.SettingForm();
                    f.SettingSaved = onClosed;
                    Application.Run(f);
                    isOpeningSettingDialog = false;
                }));
                this.uiThread.SetApartmentState(ApartmentState.STA);
                this.uiThread.Start();
            }
        }

        private void OnSettingApplying()
        {
            StopClient();
            StartClient();
        }

        private void Notify(string msg)
        {
            foreach (var inst in this.instances.Values)
            {
                // 各プラグインインスタンスへ通知
                inst.Notify(msg);
            }
        }

        private void StartClient()
        {
            this.clientStopped = new AutoResetEvent(false);
            this.clientThread = new Thread(new ThreadStart(ClientThread));
            this.clientThread.Start();
        }

        private void StopClient()
        {
            this.isClientRunning = false;
            if(null != this.clientThread)
            {
                if(this.clientThread.IsAlive)
                {
                    if(this.clientStopped.WaitOne(10000))
                    {
                        System.Diagnostics.Debug.WriteLine("clientThread stopped");
                        this.clientThread = null;
                    }
                }
            }
        }

        private void ClientThread()
        {
            this.clientFailed = new ManualResetEvent(false);
            int interval = 0;
            int failure_count = 0;

            IPEndPoint server_ep = null;
            this.client = null;
            this.isClientRunning = true;

            while(this.isClientRunning)
            {
                if(this.clientFailed.WaitOne(0))
                {
                    failure_count++;
                    if(failure_count > 10) { failure_count = 10; }

                    System.Diagnostics.Debug.WriteLine("clientFail detected, count = " + failure_count);

                    if(null != client)
                    {
                        client.Stop();
                    }
                    client = null;
                    server_ep = null;

                    // 失敗回数に応じてリトライまでの間隔を空ける
                    interval = (int)Math.Pow(2, failure_count) * 10;

                    this.clientFailed.Reset();
                }

                if (interval > 0)
                {
                    interval--;
                    Thread.Sleep(100);
                    continue;
                }
                
                if (null == server_ep)
                {
                    server_ep = GetServerEndPoint();
                    System.Diagnostics.Debug.WriteLine("got endpoint " + server_ep.ToString());
                    continue;
                }

                if (null != client)
                {
                    if (client.IsConnected)
                    {
                        failure_count = 0;
                    }
                    Thread.Sleep(100);
                    continue;
                }

                System.Diagnostics.Debug.WriteLine("try starting client ...");
                try
                {
                    client = new P2PQRelay.Client();
                    client.Address = server_ep.Address;
                    client.Port = server_ep.Port;
                    client.OnConnectionFailed += (obj, e) =>
                    {
                        System.Diagnostics.Debug.WriteLine("connection failed: " + e.Exception);
                        this.clientFailed.Set();
                    };
                    client.OnReceived += (obj, e) =>
                    {
                        System.Diagnostics.Debug.WriteLine("received: " + e.Message);

                        // P2PQRelay.Clientの持つスレッドから呼び出されているので注意
                        this.Notify(e.Message);
                    };
                    client.OnConnected += (obj, e) =>
                    {
                        System.Diagnostics.Debug.WriteLine("connected");
                    };
                    client.OnDisconnected += (obj, e) =>
                    {
                        System.Diagnostics.Debug.WriteLine("disconnected");
                        this.clientFailed.Set();
                    };
                    client.Start();
                }
                catch(Exception e)
                {
                    System.Diagnostics.Debug.WriteLine("caught exception: " + e);
                    this.clientFailed.Set();
                }
                finally
                {
                }
            }

            System.Diagnostics.Debug.WriteLine("shutting down client ...");
            client.Stop();

            // スレッドの終了をシグナル
            this.clientStopped.Set();
            return;
        }

        private IPEndPoint GetServerEndPoint()
        {
            IPEndPoint server_ep = null;
            int interval = 0;

            if (P2PQRelayClient_Net.Properties.Settings.Default.EnableAutoDetect)
            {
                while (this.isClientRunning)
                {
                    if (interval > 0)
                    {
                        interval--;
                        Thread.Sleep(100);
                        continue;
                    }

                    try
                    {
                        server_ep = P2PQRelay.Discover.Query();
                        if (null == server_ep)
                        {
                            interval = 300; // retry after 30s
                            continue;
                        }

                        P2PQRelayClient_Net.Properties.Settings.Default.CommIpAddr = server_ep.Address.ToString();
                        P2PQRelayClient_Net.Properties.Settings.Default.CommPort = server_ep.Port;
                        break;
                    }
                    catch (Exception e)
                    {
                        System.Diagnostics.Debug.WriteLine("caught exception: " + e);
                        interval = 600; // retry after 60s
                    }
                }
            }
            else
            {
                server_ep = new IPEndPoint(0, 0);
                server_ep.Address = IPAddress.Parse(P2PQRelayClient_Net.Properties.Settings.Default.CommIpAddr);
                server_ep.Port = P2PQRelayClient_Net.Properties.Settings.Default.CommPort;
            }

            return server_ep;
        }
    }
}
